#
# database.rb
#
# Copyright (c) 2003 Minero Aoki <aamine@loveruby.net>
#
# This program is free software.
# You can distribute/modify this program under the terms of
# the GNU Lesser General Public License version 2 or later.
#

require 'refe/fsdbm'
require 'refe/fileutils'
require 'refe/traceutils'

module ReFe

  class CompletionError < StandardError; end
  class LookupError < StandardError; end


  class Database

    def initialize( rootdir = nil, init = false )
      @rootdir = nil
      @init = init

      @class_document_dbm     = nil
      @method_document_dbm    = nil
      @method_csource_dbm     = nil
      @mf_relation_dbm        = nil
      @function_document_dbm  = nil
      @function_source_dbm    = nil
    end

    def class_document
      ClassTable.new(class_document_dbm())
    end

    def method_document
      MethodTable.new(method_document_dbm())
    end

    def mf_relation
      MFRelationTable.new(method_document(),
                          mf_relation_dbm(),
                          function_source())
    end

    def function_document
      FunctionTable.new(function_document_dbm())
    end

    def function_source
      FunctionTable.new(function_source_dbm())
    end

    ID_CLASS_DOCUMENT     = 'class_document'
    ID_METHOD_DOCUMENT    = 'method_document'
    ID_METHOD_CSOURCE     = 'method_csource'
    ID_MF_RELATION        = 'mf_relation'
    ID_FUNCTION_DOCUMENT  = 'function_document'
    ID_FUNCTION_SOURCE    = 'function_source'

    def class_document_dbm
      @class_document_dbm ||= FSDBM.new(dbdir(ID_CLASS_DOCUMENT))
    end

    def method_document_dbm
      @method_document_dbm ||= FSDBM.new(dbdir(ID_METHOD_DOCUMENT))
    end

    def method_csource_dbm
      @method_csource_dbm ||= FSDBM.new(dbdir(ID_METHOD_CSOURCE))
    end

    def mf_relation_dbm
      @mf_relation_dbm ||= FSDBM.new(dbdir(ID_MF_RELATION))
    end

    def function_document_dbm
      @function_document_dbm ||= FSDBM.new(dbdir(ID_FUNCTION_DOCUMENT))
    end

    def function_source_dbm
      @function_source_dbm ||= FSDBM.new(dbdir(ID_FUNCTION_SOURCE))
    end

    private

    def rootdir
      return @rootdir             if @rootdir
      return ENV['REFE_DATA_DIR'] if ENV['REFE_DATA_DIR']
      return REFE_DATA_DIR        if defined?(REFE_DATA_DIR)
      raise ArgumentError, 'ReFe database directory not given'
    end

    include ReFe::FileUtils

    def dbdir( relpath )
      dir = rootdir() + '/' + relpath
      unless File.directory?(dir)
        raise ArgumentError, "database not initialized: #{dir}" unless @init
        mkdir_p dir
      end
      dir
    end
  
  end


  class ClassTable

    include TraceUtils

    def initialize( dbm )
      @dbm = dbm
    end

    def classes
      @dbm.keys
    end

    def complete( c )
      list = @dbm.expand_to_names(c + '*')
      raise CompletionError, "not match: #{c}" if list.empty?
      return list if list.size == 1

      list.each do |n|
        return [n] if n == c
      end

      list
    end

    def []( name )
      @dbm[name] or
              raise LookupError, "class not found: #{name}"
    end

    def []=( name, content )
      @dbm[name] = content
    end

  end


  class MethodTable

    include TraceUtils

    def initialize( dbm )
      @dbm = dbm
    end

    def singleton_methods_of( c )
      @dbm.expand_to_names("#{c}./*").map {|n| n.sub(/\A[\w:]+\./, '') }
    end

    def instance_methods_of( c )
      @dbm.expand_to_names("#{c}#/*").map {|n| n.sub(/\A[\w:]+#/, '') }
    end

    TOO_MANY = 10   # number of methods which uses same entry

    def complete( c, t, m )
      list = @dbm.expand_to_names(mkpattern(c,t,m))
      trace "mtable comp/1: list.size = #{list.size}"
      raise CompletionError, "not match: #{c}#{t}#{m}" if list.empty?
      return list if list.size == 1

      alt1 = list.select {|name|
          c2, t2, m2 = name.split(/([\.\#])/, 2)
          m2 == m
      }
      trace "mtable comp/2: alt1.size = #{alt1.size}"
      return alt1 if alt1.size == 1

      alt2 = list.map {|n| n.sub(/[!?]\z/, '') }.uniq
      trace "mtable comp/3: alt2.size = #{alt2.size}"
      return alt2 if alt2.size == 1

      if list.size < TOO_MANY
        h = {}
        list.each {|n| (h[self[n]] ||= []).push n }
        trace "mtable comp/4: alt3.size = #{h.size}"
        return [h.to_a[0][1][0]] if h.size == 1
      else
        trace "mtable comp/4: alt3 skipped"
      end

      trace "mtable comp: completion failed"
      list
    end

    def []( name )
      c, t, m = name.split(/([\.\#])/, 2)
      @dbm[c + t, m] or
              raise LookupError, "method not found: #{name}"
    end

    def []=( name, content )
      c, t, m = name.split(/([\.\#])/, 2)
      @dbm[c + t, m] = content
    end

    private

    def mkpattern( c, t, m )
      any(c) + t.to_s + '/' + any(m)
    end

    def any( word )
      return '*' unless word
      word.gsub(/[a-z]/) {|ch| "[#{ch}#{ch.upcase}]" } + '*'
    end

  end


  class MFRelationTable

    def initialize( mdoc, mtof, fsrc )
      @mdoc = mdoc
      @mtof = mtof
      @fsrc = fsrc
    end

    def complete( c, t, m )
      @mdoc.complete(c,t,m)
    end

    def []( method )
      c, t, m = method.split(/([\.\#])/, 2)
      f = @mtof[c + t, m] or
              raise LookupError, "cannot convert method to function: #{method}"
      @fsrc[f]
    end

    def []=( method, function )
      c, t, m = method.split(/([\.\#])/, 2)
      @mtof[c + t, m] = function
    end

  end


  class FunctionTable
  
    include TraceUtils

    def initialize( dbm )
      @dbm = dbm
    end

    def complete( words )
      list = @dbm.expand_to_keys(mkpattern(words)).map {|a| a.join('_') }
      trace "ftable comp: #{list.inspect}" if list.size < 20
      raise CompletionError, "not match: #{words.join(' ')}" if list.empty?
      list
    end

    def mkpattern( words )
      words.join('_').tr('-', '_').squeeze('_').sub(/\A_/, 'rb_')\
              .split(/_/).map {|w| w + '*' }.join('/')
    end
    private :mkpattern

    def []( name )
      @dbm[*name.split(/_/)] or
              raise LookupError, "function not found: #{name}"
    end

    def []=( name, content )
      @dbm[*name.split(/_/)] = content
    end

  end

end
